#include <stdio.h>
#include <bios.h>
#include <conio.h>
#include <dos.h>
#include <io.h>
#include <mem.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef toupper
	#undef toupper		/* don't want a macro version with side effects */
#endif

/******************************  Warning !!!! ******************************

	This program uses writes to absolute sectors for copying and can format
	disk tracks.  If you make changes, be sure that you know what you are
	doing.  In particular, note that if a secondary floppy controller is in
	use, the drive numbers used by the absolute sector read/write routine
	are different than those used by the formatting routine.  absread and
	abswrite use A=0, B=1,... for all disks, but INT 13h floppy operations
	use this convention only for A: and B: (and maybe C: and D: if these
	are floppies).  On my system, the two floppies on my secondary controller
	are L: and M:, and as expected absread/write want drive numbers 11 and 12,
	yet they respond as 4 and 5 to INT 13h.  I think this means somebody is
	allowing for the possibility of 4 floppies on the primary controller.

****************************************************************************/

/* List of possible enhancements
	1. Insert a proper volume serial number (different from the original).
		Ignore this difference when verifying.
	2. Allow output drive != input.
	3. Configure colors.
	4. Use extended or expanded memory rather than a disk file.
	5. Configure secondary controller drive numbers.
 */

#define LOGICAL		int
#define FALSE			0
#define TRUE			1
#define START_CLOCK	0
#define READ_CLOCK	1
#define ERR				-1
#define NO_ERR			0

#define BACKGROUND_COLOR	BLUE
#define TEXT_COLOR			WHITE
#define READ_INDICATOR		RED
#define FORMAT_INDICATOR	MAGENTA
#define WRITE_INDICATOR		GREEN
#define VERIFY_INDICATOR	YELLOW

#define MAX_SECTORS	64
#define BUFFER_SIZE	MAX_SECTORS*512

#define ERROR_ALERT()	{sound(45); delay(500);	nosound();}

/* Prototypes for local routines
 */
char *getvolid(int drive);
void far hardset(void);
void measure_speed(int function, char *thing_timed);
void statline(int sectors, int color);
int disk_format(int drive);
void restore_system_dpb(void);
void erase_temp_file(void);

/* wsscopyd - copy a disk using another disk for temporary storage of
	disk image.
	usage: copyd [options] d[:] [options]

	Options (case ignored, / or - switchchar allowed):
		-B - beep when disk change is needed
		-Cn - number of copies
		-Dn - drive number for secondary controller formatting if default fails
		-F - format target disk
		-R fn - reuse saved file fn
		-S fn - save disk file with name fn
		-T - display time used
		-V - verify by reading floppy and comparing to original data

	Compile with compact or large model else allocation of buffers will fail.

	DESQView compatibility: DOS Buffers for EMS should be set larger than the
	default value of 2K else run time will be excessive.  4K is the minimum
	acceptable value, but 8K is better.  Performance continues to improve
	up to at least 32K.  The -t option can be used to time operation with
	different values.  The following read times were obtained on a 386/25:

	 EMS buffer size    Seconds
			 2K            307.8      (default setting)
			 4K            161.6
			 8K             93.5
			16K             76.3
			32K             68.4
		under DOS          59.4

	All screen output leaves a 2-character margin on left and right so that
	everything will be visible even if there is a window border at the edge
	of the screen.  The right margin also eliminates any worry about whether
	writing in the last character position of the bottom row causes the screen
	to scroll.
 */

volatile int Hard_error;	/* set to TRUE by critical error handler */
int Fail_ignore = 3;			/* tell critical error handler to fail on error */
int Tr_per_dsk;
int Sd_per_dsk;
int Sectors_per_track;
void interrupt (*System_dpb)();

struct DPB
	{
	char spec1;			/* first specify byte */
	char spec2;    	/* second specify byte */
	char stop_delay;	/* delay until motor turned off (ticks) */
	char bps;			/* bytes per sector 0=128, 1=256, 2=512, 3=1024 */
	char spt;			/* sectors per track */
	char gap;			/* gap between sectors (2Ah for 5.25", 1Bh for 3.5") */
	char data_l;		/* data length (ignored if bps is non-zero) */
	char gap_l;			/* formatting gap length (50h for 5.25", 6ch for 3.5") */
	char filler;		/* format filler byte (default F6h) */
	char settle;		/* head settle time in milliseconds */
	char start_time;	/* motor start time in 1/8 seconds */
	} my_dpb;

char Tempfile[100];		/* file name for storage disk to be copied, global
								 	so that exit routine can erase it if needed */

void main(int argc, char **argv)
	{
	char	*buffer,				/* buffer for data from floppy or save file */
			*buffer2,			/* buffer for data from floppy when verifying */
			*cmp1,				/* pointer looping through buffer */
			*cmp2,				/* pointer looping through buffer2 */
			*openmode,			/* mode for opening disk file, read or update */
			*temp_path,			/* TEMP or TMP environment variable */
			*volname;			/* volume name or <none> */
	FILE	*fp;					/* file pointer for save file */
	int	drive_no,			/* drive number, A=0, B=1,... */
			first_sec,			/* first sector to be read from floppy */
			format_drive,		/* drive number for formatting, A=0, B=1 but others
										depend on details of secondary controller */
			i,						/* for loop index */
			icopy,				/* counter for number of copies made */
			key,					/* key hit in response to a prompt */
			ncopies,				/* number of copies requested */
			nread,				/* number of sectors to read in one call */
			nremove,				/* number of removable drives between B: and
										requested drive */
			nver,					/* number of sectors to verify in one call */
			nwrite,				/* number of sectors to write in one call */
			stat,					/* status from ioctl check for removability */
			tot_sects;			/* number of sectors on source disk */
	unsigned int	j,			/* for loop index */
						nbytes;	/* number of bytes to verify on one disk read */
	LOGICAL  format,			/* if TRUE, format disks */
				have_drive,		/* set TRUE if command line specified drive */
				have_secondary,/* set TRUE if command line specified drive number
										for secondary controller formatting */
				save_temp_file,/* if TRUE, temp file is saved */
				time,				/* if TRUE, display time used by each operation */
				use_beep,		/* if TRUE, sound alarm when disk swap is needed */
				use_old_file,	/* if TRUE, use an old temp file */
				verify;			/* if TRUE, verify copy by re-reading and comparing
										with what should have been written */
	char beep;					/* control-g if alarm enabled, blank if not */
	struct BPB					/* info in floppy boot sector */
		{
		char jump[3];
		char oem[8];
		short bytes_per_sector;
		char sectors_per_alloc;
		short reserved_secs;
		char num_fats;
		short num_root_dir;
		short total_sectors;
		char media_desc;
		short sectors_per_fat;
		short sectors_per_track;
		short num_heads;
		short num_hidden;
		} *bpb;

	directvideo = 0;		/* use ROM BIOS, should avoid bleedthrough in DV
									windows on pre-386 CPU's */

	textbackground(BACKGROUND_COLOR);
	textcolor(TEXT_COLOR);
	cprintf("\r\n  WSScopyd version 1.0 beta\r\n"
			  "  Copyright July 15, 1992 "
			  "by Robert W. Babcock and WSS Division of DDC\r\n"
			  "  Unlimited non-commercial use authorized\r\n\n");

/* Need DOS 3+ for removability check
 */
	if(_osmajor < 3)
		{
		cprintf("  Sorry, WSScopyd requires DOS 3 or higher\r\n");
		exit(1);
		}

	if(argc < 2)
		{
		cprintf("  Usage: copyd [options] drive: [options]\r\n"
				 "    Options are marked by / or -\r\n"
				 "      -B - beep when disk change is needed\r\n"
				 "      -Cn - number of copies\r\n"
				 "      -Dn - drive number for secondary controller formatting if "
				 "default fails\r\n"
				 "      -F - format target disk\r\n"
				 "      -R fn - reuse saved file fn\r\n"
				 "      -S fn - save disk file with name fn\r\n"
				 "      -T - display time used\r\n"
				 "      -V - verify by reading floppy and comparing to original data\r\n");
		exit(1);
		}

	bpb = (struct BPB *)malloc(1024);
	if(NULL == (buffer = malloc(BUFFER_SIZE)))
		{
		cprintf("  Error: out of memory\r\n");
		exit(1);
		}

	temp_path = getenv("TEMP");
	if(NULL == temp_path)
		temp_path = getenv("TMP");
	if(NULL == temp_path)
		{
		cprintf("  Error: environment variable TEMP or TMP must be defined\r\n");
		exit(1);
		}

/* Parse command line
 */
	use_beep = format = time = verify = FALSE;
	have_drive = have_secondary = save_temp_file = use_old_file = FALSE;
	ncopies = 1;
	for(argc--, argv++ ; argc>0; argc--, argv++)
		{
		if('-' == **argv || '/' == **argv)
			{
			(*argv)++;
			switch(toupper(**argv))
				{
				case 'B':
					use_beep = TRUE;
					continue;

				case 'C':
					(*argv)++;
					if('\0' == **argv)
						{
						argv++;
						argc--;
						if(0 == argc)
							{
							cprintf("  Error, number of copies missing\r\n");
							exit(1);
							}
						}
					ncopies = atoi(*argv);
					continue;

				case 'D':
					(*argv)++;
					if('\0' == **argv)
						{
						argv++;
						argc--;
						if(0 == argc)
							{
							cprintf("  Error, secondary controller drive number "
										"missing\r\n");
							exit(1);
							}
						}
					format_drive = atoi(*argv);
					have_secondary = TRUE;
					continue;

				case 'F':
					format = TRUE;
					continue;

				case 'R':
					use_old_file = TRUE;
					(*argv)++;
					if('\0' == **argv)
						{
						argv++;
						argc--;
						if(0 == argc)
							{
							cprintf("  Error, old save file name missing\r\n");
							exit(1);
							}
						}
					strcpy(Tempfile, *argv);
					continue;

				case 'S':
					save_temp_file = TRUE;
					(*argv)++;
					if('\0' == **argv)
						{
						argv++;
						argc--;
						if(0 == argc)
							{
							cprintf("  Error, save file name missing\r\n");
							exit(1);
							}
						}
					strcpy(Tempfile, *argv);
					continue;

				case 'T':
					time = TRUE;
					continue;

				case 'V':
					verify = TRUE;
					continue;

				default:
					cprintf("  Unrecognized option %s, "
								"run with no arguments for help\r\n", *argv);
					exit(1);
				}
			}
		else
			{
			if(have_drive)
				{
				cprintf("  Error, missing switchchar or drive specified twice\r\n");
				exit(1);
				}
			else
				{
				drive_no = (0x1f & **argv) - 1;
				have_drive = TRUE;
				}
			}
		}

/* Check for valid options requested
 */
	if(!have_drive)
		{
		cprintf("  Error, drive not specified\r\n");
		exit(1);
		}

	if(ncopies < 0 || ((0 == ncopies) && !save_temp_file))
		{
		cprintf("  Error, requested number of copies is %d\r\n", ncopies);
		exit(1);
		}

	if(!have_secondary)
		{
		format_drive = drive_no;
		if(format && drive_no > 1)
			{
/* Drive is not A: or B:, so we have to figure out which floppy number to
	use with int 13h.  Find all removable media drives after A: and B: but
	before the requested drive.  Assume that the first drive on the secondary
	controller is number 4 for int 13h.  This algorithm fails if there is a
	device driver referencing drive A: or B: through another letter, and may
	also be dependent on what the device driver for the secondary controller
	does.
 */
			nremove = 0;
			for(i=2; i<drive_no; i++)
				{
				if(0 == ioctl(i+1, 8, 0, 0))
					nremove++;
				}
			format_drive = 4 + nremove;
			}
		}

	if(save_temp_file && use_old_file)
		{
		cprintf("  Error, -R and -S options are mutually exclusive\r\n");
		exit(1);
		}

	if(format)
		{
/*	Get and save address of old disk parameter block, copy to a replacement
	block (fields will be changed later).  Register an exit function to
	restore the old block when we terminate.
 */
		System_dpb = getvect(0x1e);
		movedata(FP_SEG(System_dpb), FP_OFF(System_dpb),
					FP_SEG(&my_dpb), FP_OFF(&my_dpb), sizeof(struct DPB));
		setvect(0x1e, (void interrupt (*)()) &my_dpb);
		atexit(restore_system_dpb);
		}

	beep = use_beep ? '\a' : ' ';

	if(verify)
		{
		if(NULL == (buffer2 = malloc(BUFFER_SIZE)))
			{
			cprintf("  Error: not enough memory for verify buffer\r\n");
			exit(1);
			}
		}

/* Make sure the disk to be copied is removable
 */
	stat = ioctl(drive_no+1, 8, 0, 0);
	if(0 != stat)
		{
		cprintf("  Error: copyd only works for removable media\r\n");
		exit(1);
		}

	if(!(save_temp_file || use_old_file))
		{
		strcpy(Tempfile, temp_path);
		if('\\' != Tempfile[strlen(Tempfile)-1])
			strcat(Tempfile, "\\");
		strcat(Tempfile, "copyd.$$$");
		}

/* Set up critical error handler.  DOS undoes this automatically when we exit.
 */
	hardset();

/* Start the main copy loop
 */
do_another:
	if(!use_old_file)
		{
		if(time)
			measure_speed(START_CLOCK, "");

		if(0 != absread(drive_no, 1, 0, bpb) || Hard_error)
			{
			ERROR_ALERT();
			cprintf("\r\n  Error: unable to read floppy boot sector\r\n");
			exit(1);
			}
		if(512 != bpb->bytes_per_sector)
			{
			ERROR_ALERT();
			cprintf("\r\n  Error: floppy boot sector indicates other than 512 "
					  "byte sectors\r\n");
			exit(1);
			}
		volname = getvolid(drive_no + 1);
		}

	openmode = use_old_file ? "rb" : "w+b";
	if(NULL == (fp = fopen(Tempfile, openmode)))
		{
		ERROR_ALERT();
		cprintf("\r\n  Error: unable to open temporary file %s\r\n", Tempfile);
		exit(1);
		}
	if(!(use_old_file || save_temp_file))
		atexit(erase_temp_file);

/* Read floppy and copy to temporary file
 */
	if(use_old_file)
		{
		if(1 != fread(bpb, sizeof(struct BPB) + 12, 1, fp)
			|| 1 != fread(buffer, 512 - 12 - sizeof(struct BPB), 1, fp)
			|| Hard_error)
			{
			cprintf("\r\n  Error reading saved file %s\r\n", Tempfile);
			exit(1);
			}
		volname = (char *)bpb + sizeof(struct BPB);
		}

	tot_sects = bpb->total_sectors;
	if(!use_old_file)
		{
		first_sec = 0;
		cprintf("\r  Reading %d sectors, Volume %s, making %d copies...",
				tot_sects, volname, ncopies);
		clreol();
		statline(-tot_sects, 0);
		memset(buffer, 0, 512);
		if(1 != fwrite(bpb, sizeof(struct BPB), 1, fp)
			|| 1 != fwrite(volname, 12, 1, fp)
			|| 1 != fwrite(buffer, 512 - 12 - sizeof(struct BPB), 1, fp)
			|| Hard_error)
			{
			ERROR_ALERT();
			cprintf("\r\n  Error: temp disk write failed\r\n");
			exit(1);
			}

		while(tot_sects > 0)
			{
			nread = min(MAX_SECTORS, tot_sects);
			if(0 != absread(drive_no, nread, first_sec, buffer) || Hard_error)
				{
				ERROR_ALERT();
				cprintf("\r\n  Error: floppy read failed\r\n");
				exit(1);
				}
			tot_sects -= nread;
			first_sec += nread;
			statline(tot_sects, READ_INDICATOR);
			if(nread != fwrite(buffer, bpb->bytes_per_sector, nread, fp))
				{
				ERROR_ALERT();
				cprintf("\r\n  Error: temp disk write failed\r\n");
				exit(1);
				}
			}

		if(time)
			measure_speed(READ_CLOCK, "Reading source disk");
		}

/* Done reading, ready to copy
 */
	for(icopy=0; icopy<ncopies; icopy++)
		{
wait_for_blank:
		cprintf("\r");		clreol();	gotoxy(1, wherey() - 1);	clreol();
		cprintf("  Waiting for blank disk, hit a key...%c", beep);
restart_after_error:
		Hard_error = FALSE;
		fseek(fp, 512, SEEK_SET);
		key = getch();
		if(3 == key || 0x1b == key)
			{
			fclose(fp);
			exit(0);
			}

		if(time)
			measure_speed(START_CLOCK, "");

		if(format)
			{
			tot_sects = bpb->total_sectors;
			cprintf("\r  Formatting %d sectors, Volume %s, copy number "
					"%d of %d...", tot_sects, volname, icopy+1, ncopies);
			clreol();
			Tr_per_dsk = tot_sects / (2 * bpb->sectors_per_track);
			Sd_per_dsk = bpb->num_heads;
			Sectors_per_track = bpb->sectors_per_track;
			statline(-2*Tr_per_dsk, 0);

			if(ERR == disk_format(format_drive))
				{
				ERROR_ALERT();
				cprintf("\r\n  Error: floppy format failed\r\n");
				goto wait_for_blank;
				}

			if(time)
				{
				measure_speed(READ_CLOCK, "Formatting disk");
				measure_speed(START_CLOCK, "");
				}
			cprintf("\r");		clreol();	gotoxy(1, wherey() - 1);
			}

		tot_sects = bpb->total_sectors;
		cprintf("\r  Writing %d sectors, Volume %s, copy number "
				"%d of %d...", tot_sects, volname, icopy+1, ncopies);
		clreol();
		first_sec = 0;
		statline(-tot_sects, 0);
		while(tot_sects > 0)
			{
			nwrite = min(MAX_SECTORS, tot_sects);
			if(nwrite != fread(buffer, bpb->bytes_per_sector, nwrite, fp))
				{
				ERROR_ALERT();
				cprintf("\r\n  Error: temp disk read failed\r\n");
				exit(1);
				}
			if(0 != abswrite(drive_no, nwrite, first_sec, buffer) || Hard_error)
				{
				ERROR_ALERT();
				cprintf("\r");		clreol();	gotoxy(1, wherey() - 1);	clreol();
				cprintf("\r\n  Error: floppy write failed, insert replacement...");
				goto restart_after_error;
				}
			tot_sects -= nwrite;
			first_sec += nwrite;
			statline(tot_sects, WRITE_INDICATOR);
			}

		if(time)
			measure_speed(READ_CLOCK, "Writing disk");

/* Optionally verify by reading floppy and temp file and comparing
 */
		if(verify)
			{
			if(time)
				measure_speed(START_CLOCK, "");
			fseek(fp, 512, SEEK_SET);
			tot_sects = bpb->total_sectors;
			cprintf("\r");		clreol();	gotoxy(1, wherey() - 1);
			cprintf("  Verifying %d sectors, Volume %s, "
					"copy number %d of %d...",
					tot_sects, volname, icopy+1, ncopies);
			clreol();
			first_sec = 0;
			statline(-tot_sects, 0);
			while(tot_sects > 0)
				{
				nver = min(MAX_SECTORS, tot_sects);
				if(nver != fread(buffer, bpb->bytes_per_sector, nver, fp))
					{
					ERROR_ALERT();
					cprintf("\r\n  Error: temp disk read failed\r\n");
					exit(1);
					}
				if(0 != absread(drive_no, nver, first_sec, buffer2) || Hard_error)
					{
					ERROR_ALERT();
					cprintf("\r\n  Error: floppy read failed during verify, "
								"insert replacement...");
					goto wait_for_blank;
					}
				nbytes = nver * bpb->bytes_per_sector;
				cmp1 = buffer;
				cmp2 = buffer2;
				for(j=0; j<nbytes; j++)
					{
					if(*cmp1++ != *cmp2++)
						{
						ERROR_ALERT();
						cprintf("\r\n  Disk verify failed with unequal compare, "
									"insert replacement...");
						exit(1);
						}
					}
				tot_sects -= nver;
				first_sec += nver;
				statline(tot_sects, VERIFY_INDICATOR);
				}
			if(time)
				cprintf("\r\n");
			}
		}

	fclose(fp);
	if(save_temp_file)
		{
		cprintf("\r\n  Disk image saved in file %s\r\n", Tempfile);
		exit(0);
		}
	if(use_old_file)
		{
		cprintf("\r\n  %d copies completed\r\n", ncopies);
		exit(0);
		}
	cprintf("\r");		clreol();	gotoxy(1, wherey() - 1);
	cprintf("  Waiting for source disk, hit a key...%c", beep);
	clreol();
	key = getch();
	if(3 == key || 0x1b == key)
		exit(0);
	goto do_another;
	}

/* statline - print initial or updated progress indicator
	Initialize with sectors = -total number of sectors on disk (or whatever is
	being counted).  On calls to update progress indicator, also check for any
	keystrokes.  Exit on ^C or escape, flush anything else.
 */
void statline(int sectors, int color)
	{
	int count, i;
	static int old_count, total_sectors;

	if(sectors < 0)
		{
		cprintf("\r\n  ");
		for(i=0; i<37; i++)
			cprintf("\304\305");		/* special chars are -- and -|- */
		cprintf("\304\264\r  ");		/* special chars are -- and -|  */
		old_count = 0;
		total_sectors = -sectors;
		}
	else
		{
		count = (int)((76L*(total_sectors - sectors))/total_sectors);
		textcolor(color);
		for(i=old_count; i<count; i++)
			cprintf("\333");			/* special char is solid block */
		textcolor(TEXT_COLOR);
		old_count = count;
		while(kbhit())
			{
			switch(getch())
				{
				case 3:			/* control-c */
				case 0x1b:		/* escape */
					cprintf("\r\n  Aborting\r\n");
					exit(1);

				default:
					continue;
				}
			}
		}
	}

/*********************************************************************
	measure_speed - measure how much time elapsed between successive calls
	Start timing by calling with function=0, end timing with function=1
	and thing_timed a descriptive string.  After a type 1 call, the start
	time is preserved so that cummulative timing can be done.
 */
#include <sys/timeb.h>
void measure_speed(int function, char *thing_timed)
	{
	static struct timeb start;
	struct timeb finish;
	long seconds;
	short ms;

	switch(function)
		{
		case 0:
			ftime(&start);
			break;

		case 1:
			ftime(&finish);
			seconds=finish.time-start.time;
			ms=finish.millitm-start.millitm;
			if(ms < 0)
				{
				seconds--;
				ms += 1000;
				}
			cprintf("\r  %s took %ld.%03hd seconds", thing_timed, seconds, ms);
			clreol();	cprintf("\r\n");
			break;
		}
	return;
	}

/* getvolid - get volume label from disk to be copied, return in static
	variable.  Arg is drive number (1=A,... not 0=A)
 */
struct dos_fcb {
	unsigned char flag;		/* ff to indicate extended fcb */
	char res[5];				/* reserved */
	unsigned char vattr;		/* attribute */
	unsigned char drive;		/* drive, 1=A, 2=B,... */
	unsigned char vn[11];	/* file or volume name */
	unsigned char fattr;		/* attributes of found file */
	char resrvd[10];			/* reserved */
	unsigned int time;		/* last write time */
	unsigned int date;		/* last write date */
	long size;					/* size of file */ 
	};
char *getvolid(int drive)
	{
	struct dos_fcb fcb1, fcb2;
	static char volid[12];
	struct SREGS sregs;
	union REGS regs;

	setdta((char far *)&fcb1);
	memset(&fcb2, 0, sizeof(struct dos_fcb));
	sregs.ds = FP_SEG(&fcb2);
	fcb2.flag = 0xff;
	fcb2.vattr = 9;
	fcb2.drive = drive;
	strncpy(fcb2.vn, "???????????", 11);
	regs.x.dx = FP_OFF(&fcb2);
	regs.h.ah = 0x11;				/* search for first entry */
	intdosx(&regs, &regs, &sregs);
	if(regs.h.al)
		strcpy(volid, "<none>");
	else
		strncpy(volid, fcb1.vn, 11);
	volid[11] = '\0';
	return(volid);
	}

/*
	Routine disk_format is derived from ZFMAT.C by Edward V. Dong.
	That code did not work for 360K in 1.2M under DOS 5, nor did it handle
	1.2M or 3.5" disks.
 */

/* ----------------------------------------------------------------

	Copyright 1988 by Edward V.  Dong, All Rights Reserved.

	This source code is placed into the public domain.  However,
	contributions are always welcome.  This is a modified form of the
	source code used in the author's ZIP program.

	Edward V Dong's Floppy Formattor is a Turbo C program to format
	floppy diskettes, based on earlier work by Frank Nystrom
   (COPYIT.C;CIS 71631,355), Kim Kokkonen (FMAT.PAS; Compuserve
   72457,2131), and Peter Norton (BOOT.ASM).  This programs formats,
   accurately, 360K diskettes in either standard 360K drives or 1.2M
	drives.

	To format 360K diskettes in 1.2 megabyte drives, the DASD must be
	set - this is only possible for DOS 3.0 or higher and for AT class
	machines.  The routines here are based on Kokkonen's Turbo Pascal
	work, ported to Turbo C.

	Edward V. Dong, 12205 Alexandria Pl, Chino, CA 91710.   18 May 1988

  ----------------------------------------------------------------*/

/*  Disk Constants...*/

#define MAX_BYTES_PER_SECTOR	512
#define MAX_SECTORS_PER_TRACK	18

/* biosdisk() BIOS function cmd calls */

#define RESET_CONTROLLER	0
#define READ_TRK				2
#define WRITE_TRK				3
#define VERIFY_TRK			4
#define FORMAT_TRK			5

/* When you initialize the disk you need to pass the bios some things.
 */

struct
	{
	char cyl,
		head,
		rec,
		num;
	} format_record[MAX_SECTORS_PER_TRACK];

/* disk_format - this does the formatting
 */
int disk_format(int drive)
	{
	register int tr, sd;
	int  s, tries, success;

	my_dpb.spt = Sectors_per_track;
	if(Tr_per_dsk < 80 || (Tr_per_dsk == 80 && Sectors_per_track == 15))
		{
		my_dpb.gap = 0x2a;		/* 5.25" disk */
		my_dpb.gap_l = 0x50;
		}
	else
		{
		my_dpb.gap = 0x1b;		/* 3.5" disk */
		my_dpb.gap_l = 0x6c;
		}

	if(0 != (0xff00 & biosdisk(0x18, drive, 0, Tr_per_dsk, Sectors_per_track,
										0, format_record)))
		{
		cprintf("  \r\nbiosdisk call with ah=18h failed\r\n");
		return(ERR);
		}

/* use BIOS to format tracks */
	for(tr = 0;   tr < Tr_per_dsk;  tr++)
		{
		for(sd = 0;  sd < Sd_per_dsk;  sd++)
			{
			for(s = 0;  s < Sectors_per_track;  s++)
				{
				format_record[s].cyl = tr;
				format_record[s].head = sd;
				format_record[s].rec = s + 1;
				format_record[s].num = 2;
				}
			success = tries = 0;
			while (!success && (tries < 3))
				{
				if(0 == biosdisk(FORMAT_TRK, drive, sd, tr, 1, Sectors_per_track,
									  &format_record[0]))
					success = 1;
				else
					{
					biosdisk(RESET_CONTROLLER, 0, 0, 0, 0, 0, (void *)0);
					tries++;
					}
				}
			if (success)
				statline(2*(Tr_per_dsk - tr) - sd, FORMAT_INDICATOR);
			else
				{
				cprintf("\r\n  Format Error - aborting...\r\n");
				exit(1);
				}
			}
		}
	return(NO_ERR);
	}

/* restore_system_dpb - called when program terminates if the pointer to the
	disk parameter block in interrupt vector 0x1e needs to be restored.  Perhaps
	doing this is unnecessary, but it seems best to leave things as we found
	them.
 */
void restore_system_dpb(void)
	{
	setvect(0x1e, System_dpb);
	biosdisk(RESET_CONTROLLER, 0, 0, 0, 0, 0, (void *)0);
	return;
	}

/* erase_temp_file - called when program terminates if the disk file used to
	hold the floppy image is to be erased.
 */
void erase_temp_file(void)
	{
	remove(Tempfile);
	return;
	}
